/*
 * bbornpv.cpp
 *
 *  Created on: Mar 10, 2018
 *      Author: joe
 */

#include <omp.h>
#include <random>
#include "bbornp.h"
#include "laplacian10thispc.h"
#include "zeropad.h"
#include "afd2dfst.h"
#include "sourcescale.h"
#include "bscat.h"
#include "scndder.h"
#include "rrec.h"
#include "waveletpad.h"

bbornp::bbornp(std::vector<axis> paxes, std::vector<axis> daxes, int srcz, int recz,
    hypercube_float *bvel, hypercube_float *src, float dtu, float fmax,
    std::vector<int> nb, std::vector<float> stimei, std::vector<int> xs) {
  /* Get axes */
  _paxes = paxes; _daxes = daxes;
  _nsx = xs.size();
  _raxes.push_back(_daxes[0]); _raxes.push_back(_daxes[1]); _raxes.push_back(_daxes[2]);
  _bvel = bvel; _src = src;

  /* Get single parameters */
  _dtu = dtu; _fmax = fmax;
  _recz = recz; _srcz = srcz;
  _nt = daxes[0].n; _nrx = daxes[1].n;
  _nz = _paxes[0].n; _nx = _paxes[1].n;
  _wfldj = 10;

  /* Check wavefield */
  if((int)_paxes.size() != 3) seperr("bbornp: needs a cube for 2D wave propagation.\n");
  /* Check data */
  if((int)_daxes.size() != 3) seperr("bbornp: data must have axes of nt, nr and ns.\n");
  /* Check background velocity */
  if(!_bvel->get_axis(1).same_axis(_paxes[0]) || !_bvel->get_axis(2).same_axis(_paxes[1])) {
    seperr("bbornp: velocity axes must be same as spatial axes of wavefield.\n");
  }
  /* Check source */
  if(_src->get_ndim() != 1) seperr("bbornp: assumes the source is a 1D wavelet.\n");
  if(!_src->get_axis(1).same_axis(daxes[0])) seperr("bbornp: source axis must be same as data.\n");


  /* Get shot times and positions */
  if(stimei.size() != xs.size()) {
    seperr("bbornp: number of shot times %d should = number of shot positions %d\n",(int)stimei.size(),(int)xs.size());
  }
  _stimei = stimei; _xsi = xs; _nb = nb;
  if(_daxes[2].n != (int)(_nb.size())) seperr("bbornp: Calculated number of shots must be same as given.\n");

  /* Create container of images */
  std::vector<axis> iaxes;
  iaxes.push_back(_paxes[0]);
  iaxes.push_back(_paxes[1]);
  iaxes.push_back(_daxes[2]);
  _imgs = new hypercube_float(iaxes,true);

  /* Create shot positions */
  int j = 0;
  for(int is = 0; is < (int)_nb.size(); ++is) {
    std::vector<int> xsi;
    std::vector<float> stimej;
    for(int ib = 0; ib < _nb[is]; ++ib) {
      xsi.push_back(_xsi[j]);
      stimej.push_back(_stimei[j]);
      ++j;
      }
    _xss.push_back(xsi);
    _stimes.push_back(stimej);
  }

}

void bbornp::setbwfldout(std::string wfname, int wfldj) {
  _bwfname = wfname; _wfldj = wfldj;
  _wfout = sregf(_paxes);
  _wfout.write_description(wfname);
}

void bbornp::setawfldout(std::string wfname, int wfldj) {
  _awfname = wfname; _wfldj = wfldj;
  _awfout = sregf(_paxes);
  _awfout.write_description(wfname);
}

bool bbornp::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornpv forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(3).same_axis(_daxes[2])) seperr("bornpv forward: data axes must be same as constructor.\n");

  /* Loop over blended shots */
  int bw = 0;
  omp_set_num_threads(_nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)_nb.size(); ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    rrec rr                 = rrec(_raxes[1],_recz);
    scndder d2              = scndder(_raxes[0]);
    waveletpad wp           = waveletpad(_src->get_axis(1));
    hypercube_float *tpd = new hypercube_float(_src->get_axes(),true);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    bscat   im              = bscat(prc);
    /* Compute background wavefield */
    if(_verb) fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    std::vector<int> xsi = _xss[is];
    std::vector<float> stimej = _stimes[is];
    for(int ib = 0; ib < _nb[is]; ++ib) {
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,_src,tpd);
      if(_verb) fprintf(stderr,"xs=%d stime=%f\n",xsi[ib],stimej[ib]);
      zeropad zp = zeropad(xsi[ib],_srcz);
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    im.forward(false,m,ssc);
    d2.forward(false,ssc,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out born wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    rr.adjoint(false,rec,prc);
    memcpy(&d->vals[is*_nt*_nrx],rec->vals,sizeof(float)*_nt*_nrx);
    delete bck; delete ssc;
    delete sht; delete rec; delete tpd;
  }

  return true;
}

bool bbornp::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornpv forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(3).same_axis(_daxes[2])) seperr("bornpv forward: data axes must be same as constructor.\n");

  /* Loop over shot */
  int bw = 0, aw = 0;
  omp_set_num_threads(_nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)_nb.size(); ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    rrec pvrr               = rrec(_raxes[1],_recz);
    scndder d2              = scndder(_raxes[0]);
    waveletpad wp           = waveletpad(_src->get_axis(1));
    hypercube_float *tpd = new hypercube_float(_src->get_axes(),true);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *img = new hypercube_float(_bvel->get_axes(),true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    bscat   im              = bscat(prc);
    /* Build blended source */
    prc->zero();
    std::vector<int> xsi = _xss[is];
    std::vector<float> stimej = _stimes[is];
    if(_verb) fprintf(stderr,"shotnum=%d\n",is);
    for(int ib = 0; ib < _nb[is]; ++ib) {
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,_src,tpd);
      if(_verb) fprintf(stderr,"xs=%d stime=%f\n",xsi[ib],stimej[ib]);
      zeropad zp = zeropad(xsi[ib],_srcz);
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    /* Compute background wavefield */
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&d->vals[_nt*_nrx*is],sizeof(float)*_nt*_nrx);
    /* Back propagate perturbation */
    pvrr.forward(false,sht,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!_awfname.empty()) {
      if(is%_wfldj == 0) {
        _awfout.write_all(_awfname,bck);
        _awfout.write_axis(_awfname,4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(false,img,bck);
    memcpy(&_imgs->vals[is*_nx*_nz],img->vals,sizeof(float)*_nx*_nz);
    delete bck; delete ssc; delete img;
    delete sht; delete rec; delete tpd;
  }

  /* Stack over all images */
  hypercube_float *timg = new hypercube_float(_bvel->get_axes(),true);
  for(int ii = 0; ii < (int)_nb.size(); ++ii) {
    memcpy(timg->vals,&_imgs->vals[ii*_nx*_nz],sizeof(float)*_nx*_nz);
    m->add(timg);
  }

  return true;
}
