/*
 * bornp.cpp
 *
 *  Created on: Mar 06, 2018
 *      Author: joe
 */

#include "bornp.h"

bornp::bornp(std::vector<axis> paxes, std::vector<axis> daxes, axis saxis, int srcz, int recz,
    hypercube_float *bvel, hypercube_float *src, float dtu, float fmax) {
  /* Get axes */
  _paxes = paxes; _daxes = daxes; _saxis = saxis;
  _nsx = _saxis.n; _osx = _saxis.o; _dsx = _saxis.d;
  _raxes.push_back(_daxes[0]); _raxes.push_back(_daxes[1]); _raxes.push_back(_daxes[2]);
  _bvel = bvel; _src = src;

  /* Get single parameters */
   _dtu = dtu; _fmax = fmax;
  _recz = recz; _srcz = srcz;
  _nt = daxes[0].n; _nrx = daxes[1].n;
  _nz = _paxes[0].n; _nx = _paxes[1].n;
  _wfldj = 10;
  _nthd = 1;

  /* Check wavefield */
  if((int)_paxes.size() != 3) seperr("bornp: needs a cube for 2D wave propagation.\n");
  /* Check data */
  if((int)_daxes.size() != 3) seperr("bornp: data must have axes of nt, nr and ns.\n");
  if(_daxes[2].n != _nsx) seperr("bornp: data must have same number of shots as shot axis.\n");
  /* Check background velocity */
  if(!_bvel->get_axis(1).same_axis(_paxes[0]) || !_bvel->get_axis(2).same_axis(_paxes[1])) {
    seperr("bornp: velocity axes must be same as spatial axes of wavefield.\n");
  }
  /* Check source */
  if(_src->get_ndim() != 1) seperr("bornp: assumes the source is a 1D wavelet.\n");
  if(!_src->get_axis(1).same_axis(daxes[0])) seperr("bornp: source axis must be same as data.\n");

  /* Create container of images */
  std::vector<axis> iaxes;
  iaxes.push_back(_paxes[0]);
  iaxes.push_back(_paxes[1]);
  iaxes.push_back(_daxes[2]);
  _imgs = new hypercube_float(iaxes,true);

}

void bornp::setbwfldout(std::string wfname, int wfldj) {
  _bwfname = wfname; _wfldj = wfldj;
  _wfout = sregf(_paxes);
  _wfout.write_description(wfname);

}

void bornp::setawfldout(std::string wfname, int wfldj) {
  _awfname = wfname; _wfldj = wfldj;
  _awfout = sregf(_paxes);
  _awfout.write_description(wfname);
}

bool bornp::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornp forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(3).same_axis(_daxes[2])) seperr("bornp forward: data axes must be same as constructor.\n");

  /* Loop over shot */
  int bw = 0;
  omp_set_num_threads(_nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < _nsx; ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    rrec rr                 = rrec(_raxes[1],_recz);
    scndder d2              = scndder(_raxes[0]);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    prc->zero(); bck->zero(); ssc->zero(); sht->zero(); rec->zero();
    bscat   im              = bscat(prc);
    /* Compute background wavefield */
    float xs = _osx + _dsx*is;
    //fprintf(stderr,"shotnum=%d shotpos=%f\n",is,xs);
    zeropad zp = zeropad(static_cast<int>(xs),_srcz);
    zp.forward(false,_src,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    im.forward(false,m,ssc);
    d2.forward(false,ssc,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out born wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    rr.adjoint(false,rec,prc);
    memcpy(&d->vals[is*_nt*_nrx],rec->vals,sizeof(float)*_nt*_nrx);
    delete bck; delete ssc;
    delete sht; delete rec;
  }

  return true;
}

bool bornp::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornp forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(3).same_axis(_daxes[2])) seperr("bornp forward: data axes must be same as constructor.\n");

  /* Loop over shot */
  int bw = 0, aw = 0;
  omp_set_num_threads(_nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < _nsx; ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    rrec rr                 = rrec(_raxes[1],_recz);
    scndder d2              = scndder(_raxes[0]);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *img = new hypercube_float(_bvel->get_axes(),true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    prc->zero(); bck->zero(); ssc->zero(); sht->zero(); rec->zero();
    bscat   im              = bscat(prc);
    /* Compute background wavefield */
    float xs = _osx + _dsx*is;
    //fprintf(stderr,"shotnum=%d shotpos=%f\n",is,xs);
    zeropad zp = zeropad(static_cast<int>(xs),_srcz);
    zp.forward(false,_src,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&d->vals[_nt*_nrx*is],sizeof(float)*_nt*_nrx);
    /* Back propagate perturbation */
    rr.forward(false,sht,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!_awfname.empty()) {
      if(is%_wfldj == 0) {
        _awfout.write_all(_awfname,bck);
        _awfout.write_axis(_awfname,4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(false,img,bck);
    memcpy(&_imgs->vals[is*_nx*_nz],img->vals,sizeof(float)*_nx*_nz);
    delete bck; delete ssc; delete img;
    delete sht; delete rec;
  }

  /* Stack over all images */
   hypercube_float *timg = new hypercube_float(_bvel->get_axes(),true);
   for(int ii = 0; ii < _nsx; ++ii) {
     memcpy(timg->vals,&_imgs->vals[ii*_nx*_nz],sizeof(float)*_nx*_nz);
     m->add(timg);
   }

  return true;
}
