# Creates blending factor, shot positions and times for 
# blended modeling and appends it to wavelet file

import os,sys, argparse, ConfigParser
import seppy
import numpy as np

def writeParamsToHeader(nb,bpos,btimes,wavfile):
  """ Writes blending factor, times and positions to the wavelet file"""
  nbstr = ""; bposstr = ""; btimesstr = ""
  for isht in range(len(nb)):
    nbstr += "%d,"%(nb[isht])
    xsi = bpos[isht]; timei = btimes[isht]
    for ib in range(nb[isht]):
      bposstr += "%d,"%(xsi[ib])
      btimesstr += "%f,"%(timei[ib])

  # Open data plane file
  assert(os.path.exists(wavfile) == True), "Wavelet file must exist before writing params."
  fp = open(wavfile,"a")
  fp.write("nb=%s\n"%(nbstr))
  fp.write("stimes=%s\n"%(btimesstr))
  fp.write("xss=%s\n"%(bposstr))
  fp.close()

# Parse the config file
conf_parser = argparse.ArgumentParser(add_help=False)
conf_parser.add_argument("-c", "--conf_file",
                         help="Specify config file", metavar="FILE")
args, remaining_argv = conf_parser.parse_known_args()
defaults = { 
    "minbld": 2,
    "maxbld": 2,
    "mindly": 0.0,
    "maxdly": 0.6,
    "space": 10,
    "seed": 1992,
    "verb": 0,
    }   
if args.conf_file:
  config = ConfigParser.SafeConfigParser()
  config.read([args.conf_file])
  defaults = dict(config.items("Defaults"))

# Parse the other arguments
# Don't surpress add_help here so it will handle -h
parser = argparse.ArgumentParser(parents=[conf_parser],description=__doc__,
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.set_defaults(**defaults)
# Input files
parser.add_argument("in=",help="Input wavelet file")
# Required arguments
requiredNamed = parser.add_argument_group('required parameters')
requiredNamed.add_argument("-nsx",help="Number of shots",required=True,type=int)
requiredNamed.add_argument("-osx",help="Shot origin (in samples)",required=True,type=int)
requiredNamed.add_argument("-dsx",help="Shot spacing (in samples)",required=True,type=int)
# Optional arguments
parser.add_argument("-minbld",help="Minimum blending factor (number of blended shots per shot)",type=int)
parser.add_argument("-maxbld",help="Maximum blending factor",type=int)
parser.add_argument("-mindly",help="Minimum shot delay (in seconds)",type=float)
parser.add_argument("-maxdly",help="Maximum shot delay (in seconds)",type=float)
parser.add_argument("-seed",help="Seed for creating times and blending factor",type=int)
parser.add_argument("-verb",help="verbosity flag",type=int)
parser.add_argument("-space",help="Spacing betwn blended shots (in samples)",required=True,type=int)
args = parser.parse_args(remaining_argv)

# Get input arguments
nsx    = args.nsx; nss = nsx
osx    = args.osx;
dsx    = args.dsx
space  = args.space
minbld = args.minbld; maxbld = args.maxbld;
mindly = args.mindly; maxdly = args.maxdly;
np.random.seed(args.seed)
verb   = args.verb

# Set up SEP
sep = seppy.sep(sys.argv)

nb = []; sumb = 0

# Create blending factor
while(nss > maxbld):
  nbld = np.random.randint(minbld,maxbld+1)
  nss -= nbld; sumb += nbld
  nb.append(nbld)

last = nsx - sumb
nb.append(last)
sumb += last
#print nb, sumb
assert(sumb == nsx), "Error: blended ranges are incorrect sum=%d."%(sumb)

# Create shot times
mindly = 0.0; maxdly = 0.6
stimei = [0.0]
for isx in range(1,nsx):
  stimei.append(np.random.rand()*maxdly)

# Non-blended shot positions
tshots = np.linspace(osx,osx+(nsx-1)*dsx,nsx).astype(int)

# Create blended shot positions
xss = []; stimes = []
# Loop over all events (blended shots)
for isx in range(len(nb)):
  xsi = []; stimej = []; idx = []; nf = 1
  # Loop over all shots in blended shot
  for ib in range(nb[isx]):
    # Save the first shot
    if(ib == 0): 
      xsi.append(tshots[0])
      stimej.append(stimei[0])
      idx.append(0)
    else:
      # Find a shot that is far enough away
      found = False;
      for k in range(len(tshots)):
        if((tshots[k] - xsi[ib-1]) == space):
          xsi.append(tshots[k])
          stimej.append(stimei[k])
          idx.append(k)
          found = True 
          break
      # If no shots are far enough, then use the end one
      if(found == False):
        xsi.append(tshots[nf])
        stimej.append(stimei[nf])
        idx.append(nf)
        nf += 1
  # Save shot locations and times
  xss.append(xsi)
  stimes.append(stimej)
  # Remove used shots point
  tshots = np.delete(tshots,idx)
  stimei = np.delete(stimei,idx)

writeParamsToHeader(nb,xss,stimes,sep.get_fname("in"))

if(verb):
  ib = 1
  for pos,time in zip(xss,stimes):
    msg = "Blended shot %d position, time"%(ib)
    print msg,pos,time
    ib += 1

