/**
 * Independent modeling of particle velocity
 * acoustic data. Shifts wavelets and then
 * models the encoded source function.
 * @author: Joseph Jennings
 * @version: 2017.05.14
 */

#include <random>
#include <omp.h>
#include "sep_params.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "afd2dfst.h"
#include "../main/pvrrec.h"
#include "../main/waveletpad.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int nthd = pars.get_int("nthrds",2);
  /* Shot axes */
  int nsx    = pars.get_int  ("nsx",  1);
  float osx  = pars.get_float("osx",0.0);
  float dsx  = pars.get_float("dsx",1.0);
  float srcz = pars.get_float("srcz",50.0);
  std::vector<float> sxs = pars.get_vfloats("sxs",10);
  std::vector<float> szs = pars.get_vfloats("szs",10);
  if(!sxs.empty() && !szs.empty()) nsx = sxs.size();
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Density at receivers */
  float rho = pars.get_float("rho",1000.0);
  /* Blending parameters */
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"blddat");
  std::vector<float> xss    = pars.get_auxvfloats("xss",100,"blddat");
  int wfldwj = pars.get_int("wfldwj",1);

  if(stimes.size() != xss.size()) {
    seperr("number of shot times %d should = number of shot positions %d\n",(int)stimes.size(),(int)xss.size());
  }

  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(2,0.0,1.0));   raxes.push_back(axis(2,0.0,1.0));
  daxes.push_back(axis(nsx,osx,dsx));

  std::vector<axis> paxes;
  /* Wavefield axes */
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  srcin.read_all("in",src);
  velin.read_all("vel",vel);

  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);

  /* Loop over shot */
  int iw = 0;
  omp_set_num_threads(nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < nsx; ++is) {
    waveletpad wp           = waveletpad(src->get_axis(1));
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    pvrrec pvrr             = pvrrec(paxes,axis(nrx,orx,drx),recz,rho);
    hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    float xs = xss[is]; 
    fprintf(stderr,"xs=%f stime=%f\n",xs,stimes[is]);
    /* Pad each shot in time */
    wp.set_time(stimes[is]);
    wp.forward(false,src,tpd);
    zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
    /* Pad in space */
    zp.forward(true,tpd,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      if(is%wfldwj == 0) {
        wfout.write_all(wfname,prc);
        wfout.write_axis(wfname,4,axis(iw+1,0,1.0,"shot"));
        ++iw;
      }
    }
    /* Multicomponent receiver restriction */
    pvrr.adjoint(false,rec,prc);
    memcpy(&dat->vals[2*is*nt*nrx],rec->vals,sizeof(float)*2*nt*nrx);
    delete prc; delete ssc; delete rec; delete tpd;
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Clean up */
  delete vel; delete dat;
}
