/**
 * Migrates encoded multicomponent acoustic data
 * Assumes that shot number is the slow axis (4)
 * and component is before that (3)
 * @author: Joseph Jennings
 * @version: 2017.05.01
 */

#include "sep_params.h"
#include "afd2dfst.h"
#include "bscat.h"
#include "sourcescale.h"
#include "scndder.h"
#include "zeropad.h"
#include "laplacian10thispc.h"
#include "../main/waveletpad.h"
#include "../main/pvrrec.h"
#include "../main/simpleplot.h"

int main(int argc, char ** argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int wfldj = pars.get_int("wfldwj",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  float rho = pars.get_float("rho",1000.0);
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"blddat");

  /* Output file for data residual */
  std::string drname = pars.get_string("rsdout","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("vel");
  sregf wfout; sregf rout;

  if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
    seperr("MigrateMCShots: input data and source wavelet must have same time axis.\n");

  if(datin.get_axis(3).n != 2)
    seperr("MigrateMCShots assumes that the component axis is the third axis.\n");

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  int   nx  = velin.get_axis(2).n;
  /* Data axes */
  int   nt  = datin.get_axis(1).n;
  int   nrx = datin.get_axis(2).n;
  int   nsx = datin.get_axis(4).n;
  float osx = datin.get_axis(4).o;
  float dsx = datin.get_axis(4).d;
  int   nc  = datin.get_axis(3).n;

  /* Build axes */
  std::vector<axis> raxes;
  raxes.push_back(datin.get_axis(1));
  raxes.push_back(datin.get_axis(2));
  raxes.push_back(datin.get_axis(3));

  std::vector<axis>paxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  std::vector<axis> iaxes;
  iaxes.push_back(velin.get_axis(1));
  iaxes.push_back(velin.get_axis(2));
  iaxes.push_back(datin.get_axis(4));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *oimg = new hypercube_float(velin.get_axes(),true);
  hypercube_float *imgs = new hypercube_float(iaxes,true);

  /* Read in input files */
  datin.read_all("in",dat);
  srcin.read_all("src",src);
  velin.read_all("vel",vel);
  if(!bwfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description("bwfldsout");
  }
  if(!awfname.empty() && !bwfname.empty()) {
    wfout.write_description("awfldsout");
  } else if(!awfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description("awfldsout");
  }
  if(!drname.empty()) {
    rout = sregf(raxes);
    rout.write_description("rsdout");
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);

  /* Loop over shot */
  int bw = 0, aw = 0;
#pragma omp parallel for default(shared)
  for(int is = 0; is < nsx; ++is) {
    waveletpad wp           = waveletpad(src->get_axis(1));
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    pvrrec pvrr             = pvrrec(paxes,raxes[1],recz,rho);
    scndder d2              = scndder(raxes[0]);
    hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *bck = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *img = new hypercube_float(velin.get_axes(),true);
    hypercube_float *sht = new hypercube_float(raxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    bscat   im              = bscat(prc);
    /* Compute background wavefield */
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    float xs = osx + dsx*is;
    fprintf(stderr,"xs=%f stime=%f\n",xs,stimes[is]);
    /* Pad the shot in time */
    wp.set_time(stimes[is]);
    wp.forward(false,src,tpd);
    zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
    /* Pad in space */
    zp.forward(false,tpd,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    fprintf(stderr,"prcmax=%f\n",prc->my_max());
    pvrr.adjoint(false,rec,prc);
    /* If user specifies, write out background wavefield */
    if(!bwfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all("bwfldsout",prc);
        wfout.write_axis("bwfldsout",4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&dat->vals[nt*nrx*nc*is],sizeof(float)*nc*nt*nrx);
    rec->scale_add(1.0,sht,-1.0);
    //fprintf(stderr,"recmax=%f\n",rec->my_max());
    if(!drname.empty()) {
      rout.write_all("rsdout",rec);
      rout.write_axis("rsdout",4,axis(is+1,0,1.0,"shot"));
    }
    /* Back propagate perturbation */
    pvrr.forward(false,rec,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!awfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all("awfldsout",bck);
        wfout.write_axis("awfldsout",4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    //fprintf(stderr,"bckmax=%f\n",bck->my_max());
    im.adjoint(false,img,bck);
    simpleplot::plot2d(img);
    memcpy(&imgs->vals[is*nx*nz],img->vals,sizeof(float)*nx*nz);
    delete bck; delete ssc; delete img;
    delete sht; delete rec;
  }

  /* Stack over all images */
  hypercube_float *timg = new hypercube_float(velin.get_axes(),true);
  for(int ii = 0; ii < nsx; ++ii) {
    memcpy(timg->vals,&imgs->vals[ii*nx*nz],sizeof(float)*nx*nz);
    oimg->add(timg);
  }
  /* Write Image */
  sregf imgout = sregf(oimg->get_axes());
  imgout.write_description("out");
  imgout.write_all("out",oimg);
  int one = 1;
  putch("n3","d",&one);
  putch("n4","d",&one);

  /* Clean up */
  delete vel; delete src; delete dat;
}
