/**
 * Linearized waveform inversion
 * with Wiener deblending filters
 * @author: Joseph Jennings
 * @version: 2017.03.11
 */

#include "sep_params.h"
#include "../main/indbornp.h"
#include "../main/pefdeblnd.h"
#include "../main/grad_taper.h"
#include "../main/cgstep.h"
#include "../main/cghelp.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  int niter    = pars.get_int("niter",10);
  int modnthd  = pars.get_int("modthrds",1);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu  = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Blending parameters */
  std::vector<int> nb       = pars.get_auxvints("nb", 100, "in");
  std::vector<int> xss      = pars.get_auxvints("xss",100,"in");
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"in");
  /* Taper parameters */
  float z1t = pars.get_float("z1t",500);
  float z2t = pars.get_float("z2t",1500);
  /* PEF parameters */
  int pefniter          = pars.get_int("pefniter",600);
  int sniniter          = pars.get_int("sniniter",300);
  float eps             = pars.get_float("eps",0.01);
  std::vector<int> j    = pars.get_vints("j",3);
  std::vector<int> nlag = pars.get_vints("nlag",3);
  int pefthrds          = pars.get_int("pefthrds",1);
  /* Check PEF parameters */
  if(nlag.empty()) {
    seperr("Please supply either lag coordinates or the filter box dimensions\n");
  }
  if(nlag.size() != 2) {
    seperr("Please supply three nlag values or a list of lag coordinates\n");
  }
  if(j.empty()) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }
  if(j.size() != 2) {
    seperr("Please supply three j values\n");
  }
  std::string m0nm = pars.get_string("m0","");
  std::string cdnm = pars.get_string("cdat","");

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("bvel");
  sregf copin = sregf("cdat");
  sregf m0in;
  if(!m0nm.empty()) {
    m0in = sregf("m0");
  }

  /* Outputs file for inversion movies */
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  std::string rmov = pars.get_string("rmov","");
  std::string dmov = pars.get_string("dmov","");
  std::string ofno = pars.get_string("ofno","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  int nt  = datin.get_axis(1).n;
  int nrx = datin.get_axis(2).n;
  int nsx = (int)xss.size();
  int ntp = pars.get_int("n1p",nt);
  int nrp = pars.get_int("n2p",nrx);
  int nz  = velin.get_axis(1).n;
  int nx  = velin.get_axis(2).n;

  std::vector<axis> paxes, iaxes, shaxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  axis saxis = axis(nsx,0.0,1.0);
  iaxes.push_back(datin.get_axis(1));
  iaxes.push_back(datin.get_axis(2));
  iaxes.push_back(saxis);

  /* Build hypercubes */
  hypercube_float *vel  = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src  = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *bdat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *mod  = new hypercube_float(velin.get_axes(),true);
  hypercube_float *idat = new hypercube_float(iaxes,true);
  hypercube_float *ddat = new hypercube_float(iaxes,true);
  hypercube_float *cdat = new hypercube_float(iaxes,true);

  /* Build inversion vectors */
  hypercube_float *res   = new hypercube_float(iaxes,true);
  hypercube_float *drs   = new hypercube_float(iaxes,true);
  hypercube_float *grd   = new hypercube_float(velin.get_axes(),true);
  hypercube_float *grdtp = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  datin.read_all("in",bdat);
  srcin.read_all("src",src);
  velin.read_all("bvel",vel);
  copin.read_all("cdat",cdat);
  if(!m0nm.empty()) {
    m0in.read_all("m0", mod);
  } else {
    mod->zero();
  }

  /* Construct independent modeling operator */
  indbornp indborn = indbornp(paxes, iaxes, srcz, recz, vel, src, dtu, fmax, stimes, xss);
  indborn.setnthrds(modnthd);

  /* Gradient taper operator */
  grad_taper tap = grad_taper(velin.get_axis(1),z1t,z2t);

  /* Wiener filter */
  pefdblnd wf = pefdblnd(pefniter, sniniter, eps, j, nlag, nb, datin.get_axes());
  wf.set_nthrds(pefthrds);
  wf.set_padding(ntp, nrp);

  /* Initialize */
  idat->zero(); res->zero();

  /* Build CG stepper */
  cgstep step = cgstep();
  step.alloc_step(mod,(norm_vector*)idat);

  /* CG output utility */
  cghelp dgn = cghelp(mmov,gmov,ofno,dmov);

  /* Conjugate gradient inversion */
  for(int iter = 0; iter < niter; ++iter) {
    /* First compute the objective function */
    fprintf(stderr,"Computing residual...\n");
    indborn.forward(false, mod, idat);
    /* First iteration operates on blended data */
    if(iter == 0) {
      memcpy(ddat->vals,cdat->vals,sizeof(float)*nt*nrx*nsx);
    } else {
      fprintf(stderr,"Apply deblending filter...\n");
      wf.apply(bdat,idat,ddat);
    }
    res->scale_add(1.0,idat,-1.0,ddat);
    double f0 = (0.5)*(res->dot(res));
    /* Compute the gradient*/
    fprintf(stderr,"Computing gradient...\n");
    indborn.adjoint(false, grd, res);
    if(z1t != 0.0 || z2t != 0.0) {
      /* Taper the gradient */
      tap.forward(false,grd,grdtp);
    } else {
      memcpy(grdtp->vals,grd->vals,sizeof(float)*nz*nx);
    }
    /* Compute data space gradient */
    fprintf(stderr,"Computing data-space gradient...\n");
    indborn.forward(false, grdtp, drs);
    /* Compute the step length and update*/
    if(!step.step_it(mod,grdtp,res,drs)) break;
    /* Reevaluate the objective function and output */
    double f1 = (0.5)*res->dot(res);
    if(f1 >= f0) {
      fprintf(stderr, "Objective function did not reduce, will terminate solver.\n");
      break;
    }
    dgn.output(res,ddat,grdtp,mod,f1);
    fprintf(stderr,"iter=\t%d\t obj=\t%g\n",iter+1,f1);
  }

  /* Write out inverted model */
  sregf imodout = sregf(mod->get_axes());
  imodout.write_description("out");
  imodout.write_all("out",mod);
  int one = 1;
  putch("n3","d",&one);
  putch("n4","d",&one);

}
