/**
 * Blended modeling of multicomponent
 * acoustic data. Blends wavelets and then
 * models the blended source function.
 * @author: Joseph Jennings
 * @version: 2017.04.23
 */

#include <random>
#include "sep_params.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "afd2dfst.h"
#include "../main/mcrrec.h"
#include "../main/waveletpad.h"
#include "../main/blendnc2d.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  /* Shot axes */
  int nsx    = pars.get_int  ("nsx",  1);
  float osx  = pars.get_float("osx",0.0);
  float dsx  = pars.get_float("dsx",1.0);
  float srcz = pars.get_float("srcz",50.0);
  std::vector<float> sxs = pars.get_vfloats("sxs",10);
  std::vector<float> szs = pars.get_vfloats("szs",10);
  if(!sxs.empty() && !szs.empty()) nsx = sxs.size();
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Density at receivers */
  float rho = pars.get_float("rho",1000.0);
  /* Blending parameters */
  int minblnd  = pars.get_int("minblnd",2);
  int maxblnd  = pars.get_int("maxblnd",3);
  float maxdly = pars.get_float("maxdly",0.5);
  float mindly = pars.get_float("mindly",0.0);
  int seed     = pars.get_int("seed",0);
  std::string enctag = pars.get_string("encsrcs","");
  std::vector<float> stimes = pars.get_vfloats("stimes",100);
  int wfldwj = pars.get_int("wfldwj",1);

  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Build random shot times if none provided*/
  if(stimes.empty()) {
    std::random_device rd; std::mt19937 rng;
    if(!seed) {
      rng.seed(rd());
    } else {
      rng.seed(seed);
    }
    std::uniform_real_distribution<float> unif(maxdly,mindly);
    /* First time is always zero */
    stimes.push_back(0.0);
    for(int is = 1; is < nsx; ++is) {
      auto val = unif(rng);
      stimes.push_back(val);
    }
  } else {
    if((int)stimes.size() != nsx) {
      seperr("Number of shot times %d must equal number of shots %d\n",(int)stimes.size(),nsx);
    }
  }

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(nsx,minblnd,maxblnd);
  std::vector<int> nb = gam->get_nb();

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nz = velin.get_axis(1).n;
  int   nx = velin.get_axis(2).n;
  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(3,0.0,1.0));   raxes.push_back(axis(3,0.0,1.0));
  daxes.push_back(axis((int)nb.size(),0.0,1.0));

  std::vector<axis> paxes;
  /* Wavefield axes */
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *prc = new hypercube_float(paxes,true);
  hypercube_float *ssc = new hypercube_float(paxes,true);
  hypercube_float *rec = new hypercube_float(raxes,true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  std::vector<int> nwind,fwind;
  nwind.push_back(nz); nwind.push_back(nx);
  nwind.push_back(nt); nwind.push_back(1);
  fwind.push_back(0); fwind.push_back(0);
  fwind.push_back(0); fwind.push_back(0);
  srcin.read_all("in",src);
  velin.read_all("vel",vel);

  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);
  laplacian10thispc *lp   = new laplacian10thispc(paxes);
  afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
  mcrrec mcrr             = mcrrec(paxes,axis(nrx,orx,drx),recz, rho);
  waveletpad wp           = waveletpad(src->get_axis(1));

  /* Loop over shot */
  int iw = 0; int j = 0;
  for(int is = 0; is < (int)nb.size(); ++is) {
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    for(int ib = 0; ib < nb[is]; ++ib) {
      float xs = osx + dsx*j;
      /* Pad each shot in time */
      wp.set_time(stimes[j]);
      wp.forward(false,src,tpd);
      zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
      /* Blend the shots */
      zp.forward(true,tpd,prc);
      ++j;
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      if(is%wfldwj == 0) {
        wfout.write_all(wfname,prc);
        wfout.write_axis(wfname,4,axis(iw+1,0,1.0,"shot"));
        ++iw;
      }
    }
    /* Multicomponent receiver restriction */
    mcrr.adjoint(false,rec,prc);
    memcpy(&dat->vals[3*is*nt*nrx],rec->vals,sizeof(float)*3*nt*nrx);
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Write out shot times and positions */
  printf("stimes=");
  for(int is = 0; is < (int)stimes.size(); ++is) {
    printf("%f,",stimes[is]);
  }
  printf("\n");
  printf("nsx=%d\n",nsx);
  printf("osx=%f\n",osx);
  printf("dsx=%f\n",dsx);
  printf("maxblnd=%d\n",maxblnd);
  printf("minblnd=%d\n",minblnd);

  /* Clean up */
  delete vel; delete prc; delete dat; delete rec;
  delete lp;
}
