/**
 * Blended modeling of pressure data.
 * Blends wavelets and then
 * models the blended source function.
 * @author: Joseph Jennings
 * @version: 2018.03.10
 */

#include <omp.h>
#include "sep_params.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "laplacian10thispc.h"
#include "afd2dfst.h"
#include "rrec.h"
#include "../main/waveletpad.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int nthd = pars.get_int("nthrds",2);
  /* Shot axes */
  float srcz = pars.get_float("srcz",50.0);
  std::vector<float> sxs = pars.get_vfloats("sxs",10);
  std::vector<float> szs = pars.get_vfloats("szs",10);
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Blending parameters */
  std::vector<float> stimel = pars.get_auxvfloats("stimes",100,"in");
  std::vector<int> nb       = pars.get_auxvints("nb",100,"in");
  std::vector<int> xsl      = pars.get_auxvints("xss",100,"in");
  int wfldwj = pars.get_int("wfldwj",1);


  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis((int)nb.size(),0.0,1.0));

  std::vector<axis> paxes;
  /* Wavefield axes */
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  srcin.read_all("in",src);
  velin.read_all("vel",vel);

  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);

  /* Create shot positions */
  std::vector<std::vector<int>> xss;
  std::vector<std::vector<float>> stimes;
  /* Create shot positions */
  int j = 0;
  for(int is = 0; is < (int)nb.size(); ++is) {
    std::vector<int> xsi;
    std::vector<float> stimej;
    for(int ib = 0; ib < nb[is]; ++ib) {
      xsi.push_back(xsl[j]);
      stimej.push_back(stimel[j]);
      ++j;
    }
    xss.push_back(xsi);
    stimes.push_back(stimej);
  }

  /* Loop over shot */
  int iw = 0;
  omp_set_num_threads(nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)nb.size(); ++is) {
    waveletpad wp           = waveletpad(src->get_axis(1));
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    rrec rr                 = rrec(axis(nrx,orx,drx),recz);
    hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    std::vector<int> xsi = xss[is];
    std::vector<float> stimej = stimes[is];
    for(int ib = 0; ib < nb[is]; ++ib) {
      fprintf(stderr,"xs=%d stime=%f\n",xsi[ib],stimej[ib]);
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,src,tpd);
      zeropad zp = zeropad(xsi[ib],static_cast<int>(srcz));
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      if(is%wfldwj == 0) {
        wfout.write_all(wfname,prc);
        wfout.write_axis(wfname,4,axis(iw+1,0,1.0,"shot"));
        ++iw;
      }
    }
    /* Receiver restriction */
    rr.adjoint(false,rec,prc);
    memcpy(&dat->vals[is*nt*nrx],rec->vals,sizeof(float)*nt*nrx);
    delete prc; delete ssc; delete rec; delete tpd;
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Write out blending factor, shot times and positions */
  std::ostringstream nbstream, ststream, postream;
  for(int is = 0; is < (int)nb.size(); ++is) {
    nbstream << nb[is] << ",";
    std::vector<float> stimej = stimes[is];
    std::vector<int> xsi = xss[is];
    for(int ib = 0; ib < nb[is]; ++ib) {
      ststream << stimej[ib] << ",";
      postream << xsi[ib] << ",";
    }
  }
  std::cout << "nb=" << nbstream.str() << std::endl;
  std::cout << "stimes=" << ststream.str() << std::endl;
  std::cout << "xss=" << postream.str() << std::endl;

  /* Clean up */
  delete vel; delete dat;
}
