/**
 * Migrates blended particle velocity acoustic data.
 * Assumes that shot number is the slow axis (4)
 * and component is before that (3)
 * @author: Joseph Jennings
 * @version: 2017.04.25
 */

#include "sep_params.h"
#include "afd2dfst.h"
#include "bscat.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "scndder.h"
#include "rrec.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "../main/pvrrec.h"
#include "../main/waveletpad.h"
#include "../main/blendnc2d.h"
#include "../main/simpleplot.h"

int main(int argc, char ** argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int wfldj = pars.get_int("wfldwj",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  float rho = pars.get_float("rho",1000.0);
  /* Blending parameters */
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"in");
  std::vector<int> maxblnd = pars.get_auxvints("maxblnd",1,"in");
  std::vector<int> minblnd = pars.get_auxvints("maxblnd",1,"in");

  /* Output file for data residual */
  std::string drname = pars.get_string("rsdout","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("vel");
  sregf wfout; sregf rout;

  if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
    seperr("MigBldMCShots: input data and source wavelet must have same time axis.\n");

  if(datin.get_axis(3).n != 2)
    seperr("MigBldMCShots assumes that the component axis is the third axis.\n");

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  int   nx  = velin.get_axis(2).n;
  /* Data axes */
  int   nt  = datin.get_axis(1).n;
  int   nrx = datin.get_axis(2).n;
  int   nc  = datin.get_axis(3).n;
  std::vector<int> nsx = pars.get_auxvints("nsx",1,"in");
  std::vector<float> osx = pars.get_auxvfloats("osx",1,"in");
  std::vector<float> dsx = pars.get_auxvfloats("dsx",1,"in");

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(nsx[0],minblnd[0],maxblnd[0]);
  std::vector<int> nb = gam->get_nb();

  /* Build axes */
  std::vector<axis> raxes;
  raxes.push_back(datin.get_axis(1));
  raxes.push_back(datin.get_axis(2));
  raxes.push_back(datin.get_axis(3));

  std::vector<axis>paxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *prc = new hypercube_float(paxes,true);
  hypercube_float *bck = new hypercube_float(paxes,true);
  hypercube_float *ssc = new hypercube_float(paxes,true);
  hypercube_float *sht = new hypercube_float(raxes,true);
  hypercube_float *rec = new hypercube_float(raxes,true);
  hypercube_float *dat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *img = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  std::vector<int> nwind,fwind;
  nwind.push_back(nz); nwind.push_back(nx);
  nwind.push_back(nt); nwind.push_back(1);
  fwind.push_back(0); fwind.push_back(0);
  fwind.push_back(0); fwind.push_back(0);
  datin.read_all("in",dat);
  velin.read_all("vel",vel);
  srcin.read_all("src",src);
  if(!bwfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(bwfname);
  }
  if(!awfname.empty() && !bwfname.empty()) {
    wfout.write_description(awfname);
  } else if(!awfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(awfname);
  }
  if(!drname.empty()) {
    rout = sregf(rec->get_axes());
    rout.write_description("rsdout");
  }


  /* Construct operators */
  waveletpad wp           = waveletpad(src->get_axis(1));
  sourcescale ss          = sourcescale(dtu,vel);
  laplacian10thispc *lp   = new laplacian10thispc(paxes);
  afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
  pvrrec pvrr             = pvrrec(paxes,datin.get_axis(2),recz,rho);
  rrec rr                 = rrec(datin.get_axis(2),recz);
  scndder d2              = scndder(datin.get_axis(1));
  bscat   im              = bscat(prc);

  /* Loop over shot */
  int bw = 0, aw = 0, j = 0;
  for(int is = 0; is < (int)nb.size(); ++is) {
    /* Compute background wavefield */
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    prc->zero();
    for(int ib = 0; ib < nb[is]; ++ib) {
      float xs = osx[0] + dsx[0]*j;
      /* Pad each shot in time */
      wp.set_time(stimes[j]);
      wp.forward(false,src,tpd);
      zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
      /* Blend the shots */
      zp.forward(true,tpd,prc);
      ++j;
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    pvrr.adjoint(false,rec,prc);
    /* If user specifies, write out background wavefield */
    if(!bwfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all(bwfname,prc);
        wfout.write_axis(bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&dat->vals[nt*nrx*nc*is],sizeof(float)*nc*nt*nrx);
    rec->scale_add(1.0,sht,-1.0);
    if(!drname.empty()) {
      rout.write_all("rsdout",rec);
      rout.write_axis("rsdout",4,axis(is+1,0,1.0,"shot"));
    }
    /* Back propagate perturbation */
    pvrr.forward(false,rec,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!awfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all(awfname,bck);
        wfout.write_axis(awfname,4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(true,img,bck);
  }

  /* Write Image */
  sregf imgout = sregf(img->get_axes());
  imgout.write_description("out");
  imgout.write_all("out",img);
  imgout.write_axis("out",4,axis(1,0.0,1.0));

  /* Clean up */
  delete vel; delete src; delete dat;
  delete lp; delete img; delete bck; delete sht;
  delete rec; delete tpd;
}
