/**
 * Migrates blended pressure data.
 * Assumes that shot number is the slow axis (4)
 * @author: Joseph Jennings
 * @version: 2018.03.10
 */

#include <omp.h>
#include "sep_params.h"
#include "afd2dfst.h"
#include "bscat.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "scndder.h"
#include "rrec.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "rrec.h"
#include "../main/waveletpad.h"
#include "../main/simpleplot.h"

int main(int argc, char ** argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int nthd = pars.get_int("nthrds",2);
  int wfldj = pars.get_int("wfldwj",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Blending parameters */
  std::vector<float> stimel = pars.get_auxvfloats("stimes",100,"in");
  std::vector<int> nb       = pars.get_auxvints("nb",100,"in");
  std::vector<int> xsl      = pars.get_auxvints("xss",100,"in");

  /* Output file for data residual */
  std::string drname = pars.get_string("rsdout","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("vel");
  sregf wfout; sregf rout;

  if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
    seperr("MigBldShots: input data and source wavelet must have same time axis.\n");

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  int   nx  = velin.get_axis(2).n;
  /* Data axes */
  int   nt  = datin.get_axis(1).n;
  int   nrx = datin.get_axis(2).n;

  /* Build axes */
  std::vector<axis> raxes;
  raxes.push_back(datin.get_axis(1));
  raxes.push_back(datin.get_axis(2));

  std::vector<axis>paxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  std::vector<axis> iaxes;
  iaxes.push_back(velin.get_axis(1));
  iaxes.push_back(velin.get_axis(2));
  iaxes.push_back(datin.get_axis(3));


  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *oimg = new hypercube_float(velin.get_axes(),true);
  hypercube_float *imgs = new hypercube_float(iaxes,true);

  /* Read in input files */
  datin.read_all("in",dat);
  velin.read_all("vel",vel);
  srcin.read_all("src",src);
  if(!bwfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(bwfname);
  }
  if(!awfname.empty() && !bwfname.empty()) {
    wfout.write_description(awfname);
  } else if(!awfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(awfname);
  }
  if(!drname.empty()) {
    rout = sregf(raxes);
    rout.write_description("rsdout");
  }

  /* Create shot positions */
  std::vector<std::vector<int>> xss;
  std::vector<std::vector<float>> stimes;
  /* Create shot positions */
  int j = 0;
  for(int is = 0; is < (int)nb.size(); ++is) {
    std::vector<int> xsi;
    std::vector<float> stimej;
    for(int ib = 0; ib < nb[is]; ++ib) {
      xsi.push_back(xsl[j]);
      stimej.push_back(stimel[j]);
      ++j;
    }
    xss.push_back(xsi);
    stimes.push_back(stimej);
  }

  /* Construct operators */
  sourcescale ss = sourcescale(dtu,vel);

  /* Loop over shot */
  int bw = 0, aw = 0;
  omp_set_num_threads(nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)nb.size(); ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    rrec rr                 = rrec(raxes[1],recz);
    scndder d2              = scndder(raxes[0]);
    waveletpad wp           = waveletpad(src->get_axis(1));
    hypercube_float *tpd = new hypercube_float(src->get_axes(),true);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *bck = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *img = new hypercube_float(vel->get_axes(),true);
    hypercube_float *sht = new hypercube_float(raxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    bscat   im              = bscat(prc);
    /* Build blended source */
    prc->zero();
    std::vector<int> xsi = xss[is];
    std::vector<float> stimej = stimes[is];
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    for(int ib = 0; ib < nb[is]; ++ib) {
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,src,tpd);
      if(verb) fprintf(stderr,"xs=%d stime=%f\n",xsi[ib],stimej[ib]);
      zeropad zp = zeropad(xsi[ib],srcz);
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    /* Compute background wavefield */
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    rr.adjoint(false,rec,prc);
    /* If user specifies, write out background wavefield */
    if(!bwfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all(bwfname,prc);
        wfout.write_axis(bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&dat->vals[nt*nrx*is],sizeof(float)*nt*nrx);
    rec->scale_add(1.0,sht,-1.0);
    if(!drname.empty()) {
      rout.write_all("rsdout",rec);
      rout.write_axis("rsdout",3,axis(is+1,0,1.0,"shot"));
    }
    /* Back propagate perturbation */
    rr.forward(false,rec,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!awfname.empty()) {
      if(is%wfldj == 0) {
        wfout.write_all(awfname,bck);
        wfout.write_axis(awfname,4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(false,img,bck);
    memcpy(&imgs->vals[is*nx*nz],img->vals,sizeof(float)*nx*nz);
    delete bck; delete ssc; delete img;
    delete sht; delete rec; delete tpd;
  }

  /* Stack over all images */
  hypercube_float *timg = new hypercube_float(vel->get_axes(),true);
  for(int ii = 0; ii < (int)nb.size(); ++ii) {
    memcpy(timg->vals,&imgs->vals[ii*nx*nz],sizeof(float)*nx*nz);
    oimg->add(timg);
  }

  /* Write Image */
  sregf imgout = sregf(oimg->get_axes());
  imgout.write_description("out");
  imgout.write_all("out",oimg);
  imgout.write_axis("out",3,axis(1,0.0,1.0));
  imgout.write_axis("out",4,axis(1,0.0,1.0));

  /* Clean up */
  delete vel; delete src; delete dat;
  delete oimg;
}
