/**
 * Deblends 2D simultaneous source data using
 * multicomponent PEFs estimated on
 * independently modeled particle velocity data
 * @author: Joseph Jennings
 * @version: 2017.04.30
 */

#include "sep_params.h"
#include "combo_oper.h"
#include "../main/lv_conv3df.h"
#include "../main/lv_conv3dmask.h"
#include "../main/identity.h"
#include "../main/pefsignoi3d.h"
#include "../main/conj_grad.h"
#include "../main/blendnc2d.h"
#include "../main/pad3d.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  /* PEF Parameters */
  int pefniter = pars.get_int("pefniter",600);
  int sniniter = pars.get_int("sniniter",300);
  int vrb = pars.get_int("verb",0);
  float eps = pars.get_float("eps",0.01);
  float shtscale = pars.get_float("scale",10000);
  std::vector<int> j = pars.get_vints("j",3);
  std::vector<int> nlag = pars.get_vints("nlag",3);
  /* Check PEF parameters */
  if(nlag.empty()) {
    seperr("Please supply either lag coordinates or the filter box dimensions\n");
  }
  if(nlag.size() != 3) {
    seperr("Please supply three nlag values or a list of lag coordinates\n");
  }
  if(j.empty()) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }
  if(j.size() != 3) {
    seperr("Please supply three j values\n");
  }
  /* Blending parameters */
  std::vector<int> maxblnd = pars.get_auxvints("maxblnd",1,"in");
  std::vector<int> minblnd = pars.get_auxvints("maxblnd",1,"in");
  std::vector<int> nsxv = pars.get_auxvints("nsx",1,"in");
  int nsx = nsxv[0];

  /* Input data */
  sregf bldin = sregf("in");
  sregf indin = sregf("inddat");
  if(nsx != indin.get_axis(4).n) {
    seperr("Number of unblended shots from unblended and blended data is not consistent"
        " %d %d.\n",nsx,indin.get_axis(4).n);
  }

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(nsx,minblnd[0],maxblnd[0]);
  std::vector<int> nb = gam->get_nb();

  /* Get tags for movies */
  std::vector<std::string> rmovtags;
  for(int is = 0; is < nsx; ++is) {
    std::string num = std::to_string(is);
    rmovtags.push_back(pars.get_string("rsd"+num,""));
  }
  std::vector<std::string> smmvtags, nmmvtags;
  std::vector<std::string> wstags, wntags;
  for(int ib = 0; ib < (int)nb.size(); ++ib) {
    std::string num = std::to_string(ib);
    smmvtags.push_back(pars.get_string("smmv"+num,""));
    nmmvtags.push_back(pars.get_string("nmmv"+num,""));
    wstags.push_back(pars.get_string("ws"+num,""));
    wntags.push_back(pars.get_string("wn"+num,""));
  }

  /* Data axes */
  int nt   = indin.get_axis(1).n; int nrx   = indin.get_axis(2).n; int nc   = indin.get_axis(3).n;
  float ot = indin.get_axis(1).o; float orx = indin.get_axis(2).o; float oc = indin.get_axis(3).o;
  float dt = indin.get_axis(1).d; float drx = indin.get_axis(2).d; float dc = indin.get_axis(3).d;
  std::vector<int> nd;
  /* Padded axes */
  int ntp = pars.get_int("n1p",nt);
  int nrp = pars.get_int("n2p",nrx);
  int ncp = pars.get_int("n3p",3);
  nd.push_back(ntp); nd.push_back(nrp); nd.push_back(ncp);
  if(ntp == nt || nrp == nrx) {
    seperr("You have not padded the time or receiver axes. Please do so and rerun.\n");
  }
  pad3d pad = pad3d(0,0,0,0);

  /* Build axes */
  std::vector<axis> saxes, paxes;
  saxes.push_back(bldin.get_axis(1));
  saxes.push_back(bldin.get_axis(2));
  saxes.push_back(bldin.get_axis(3));
  paxes.push_back(axis(ntp,ot,dt));
  paxes.push_back(axis(nrp,orx,drx));
  paxes.push_back(axis(ncp,oc,dc));

  /* Build hypercubes */
  hypercube_float *blddat = new hypercube_float(bldin.get_axes(),true);
  hypercube_float *inddat = new hypercube_float(indin.get_axes(),true);
  hypercube_float *datsep = new hypercube_float(indin.get_axes(),true);

  /* Output data container */
  std::vector<std::vector<hypercube_float*>> odatcont((int)nb.size(),std::vector<hypercube_float*>(2));

  /* Output weighted data */
  sregf wdatout = sregf(paxes);

  /* Read in the data */
  bldin.read_all("in",blddat);
  indin.read_all("inddat",inddat);

  /* Create shot container */
  int k = 0;
  std::vector<std::vector<hypercube_float*>> shts;
  std::vector<std::vector<std::tuple<std::string,int>>> rmovtagsord;
  for(int is = 0; is < (int)nb.size(); ++is) {
    std::vector<hypercube_float*> isht;
    std::vector<std::tuple<std::string,int>> irmov;
    for(int ib = 0; ib < nb[is]; ++ib) {
      hypercube_float *tsht = new hypercube_float(saxes,true);
      memcpy(tsht->vals,&inddat->vals[k*nt*nrx*nc],sizeof(float)*nt*nrx*nc);
      isht.push_back(tsht);
      irmov.push_back(std::make_tuple(rmovtags[k],k));
      k++;
    }
    shts.push_back(isht);
    rmovtagsord.push_back(irmov);
  }

  /* Loop over all shots */
//#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)nb.size(); ++is) {
    hypercube_float *psht = new hypercube_float(paxes,true);
    hypercube_float *idat = new hypercube_float(paxes,true);
    hypercube_float *bldsht = new hypercube_float(saxes,true);
    std::vector<hypercube_float*> isht = shts[is];
    std::vector<hypercube_float*> flts;
    std::vector<std::tuple<std::string,int>> irmov = rmovtagsord[is];
    std::vector<lv_conv3dm*> pefops;
    fprintf(stderr,"blended shot %d\n",is);
    if(nb[is] < 2) {
      std::vector<hypercube_float*> outs;
      memcpy(bldsht->vals,&blddat->vals[is*nt*nrx*nc],sizeof(float)*nt*nrx*nc);
      outs.push_back(bldsht);
      odatcont[is] = outs;
      continue;
    }
    /* Estimate a PEF for each indpendent shot */
    for(int ib = 0; ib < nb[is]; ++ib) {
      fprintf(stderr,"ind shot %d\n",ib);
      /* Get and pad the shot */
      hypercube_float *ssht = isht[ib];
      ssht->scale(shtscale);
      pad.forward(false,ssht,psht);

      /* Create convolution operator */
      lv_conv3df *lvf = new lv_conv3df(nd,j,nlag,psht);

      /* Build model and data vectors */
      lvf->set_data(idat); idat->scale(-1.0);
      flts.push_back(new hypercube_float(lvf->get_filtaxes(),true));
      flts[ib]->zero(); lvf->set_zerolag(flts[ib]);
      hypercube_float *zro = new hypercube_float(lvf->get_filtaxes(),true);
      zro->zero();

      /* Create masked operator */
      lv_conv3dmask *k = new lv_conv3dmask();
      lvf->set_domain(flts[ib]); lvf->set_range(idat);
      k->set_domain(flts[ib]); k->set_range(flts[ib]);
      chain_op *dk = new chain_op(k,lvf);

      /* Create the identity operator */
      identity *iop = new identity();
      iop->set_domain(flts[ib]); iop->set_range(flts[ib]);

      /* Run the regularized inversion */
      conj_grad *cg = new conj_grad(pefniter,false);
      if(!std::get<0>(irmov[ib]).empty()) {
        cg->set_resout("rsd"+std::to_string(std::get<1>(irmov[ib])));
      }
      cg->run(dk,iop,idat,zro,flts[ib],eps);
      lvf->set_zerolag(flts[ib]);
      pefops.push_back(new lv_conv3dm(nd,j,nlag,flts[ib]));
      /* Clean up */
      delete ssht; delete lvf;
    }
    fprintf(stderr,"blended shot %d signal and noise inversion.\n",is);
    /* Get scaled and padded blended shot */
    memcpy(bldsht->vals,&blddat->vals[is*nt*nrx*nc],sizeof(float)*nt*nrx*nc);
    hypercube_float *pbldsht = new hypercube_float(paxes,true);
    pad.forward(false,bldsht,pbldsht);
    pefsignoi3d *A = new pefsignoi3d(pbldsht,pefops[0],pefops[1]);

    /* Create weighted data */
    hypercube_float *wdats = new hypercube_float(paxes,true);
    hypercube_float *wdatn = new hypercube_float(paxes,true);
    pefops[0]->forward(false,pbldsht,wdats);
    pefops[1]->forward(false,pbldsht,wdatn);

    /* Write out weighted data */
    if(!wstags[is].empty()) {
      std::string tag = "ws"+std::to_string(is);
      wdatout.write_description(tag);
      wdatout.write_all(tag,wdats);
    }
    if(!wntags[is].empty()) {
      std::string tag = "wn"+std::to_string(is);
      wdatout.write_description(tag);
      wdatout.write_all(tag,wdatn);
    }

    /* Total data vector */
    super_vector *wdat   = new super_vector(wdatn,wdats);
    super_vector *datinv = new super_vector(pbldsht,wdat);

    /* Multi-model vector */
    hypercube_float *sig  = new hypercube_float(paxes,true);
    hypercube_float *noiz = new hypercube_float(paxes,true);
    super_vector *sn = new super_vector(sig,noiz);

    /* Signal and noise inversion */
    conj_grad cg = conj_grad(sniniter);
    if(!smmvtags[is].empty()) {
      cg.set_mod1out("smmv"+std::to_string(is));
    }
    if(!nmmvtags[is].empty()) {
      cg.set_mod2out("nmmv"+std::to_string(is));
    }
    cg.run(A,datinv,sn);

    /* Save output */
    std::vector<hypercube_float*> outs;
    outs.push_back((hypercube_float*)sn->vecs[0]);
    outs.push_back((hypercube_float*)sn->vecs[1]);
    odatcont[is] = outs;

    /* Clean up */
    delete psht; delete idat;
    delete bldsht; delete pbldsht;
    delete wdats; delete wdatn;
    //    delete wdat; delete datinv;
    //    for(int ib = 0; ib < nb[is]; ++ib) {
    //      delete flts[ib]; delete pefops[ib];
    //    }
    //delete A;
  }

  /* Unwrap data container */
  k = 0;
  for(int is = 0; is < (int)nb.size(); ++is) {
    fprintf(stderr,"is=%d\n",is);
    std::vector<hypercube_float*> odat = odatcont[is];
    for(int ib = 0; ib < nb[is]; ++ib) {
      hypercube_float *tdat = new hypercube_float(saxes,true);
      pad.adjoint(false,tdat,odat[ib]);
      tdat->scale(1/shtscale);
      memcpy(&datsep->vals[k*nt*nrx*nc],tdat->vals,sizeof(float)*nt*nrx*nc);
      k++;
    }
  }

  /* Write out data */
  sregf datout = sregf(datsep->get_axes());
  datout.write_description("out");
  datout.write_all("out",datsep);

  /* Clean up */
  delete blddat; delete inddat; delete datsep;
}
