/**
 * Deblends 2D simultaneous source data using
 * PEFs estimated on independently modeled data
 * @author: Joseph Jennings
 * @version: 2018.03.11
 */

#include "sep_params.h"
#include "../main/pefdeblnd.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  /* PEF Parameters */
  int pefniter = pars.get_int("pefniter",1000);
  int sniniter = pars.get_int("sniniter",300);
  int vrb      = pars.get_int("verb",1);
  int nthd     = pars.get_int("nthrds",1);
  float eps    = pars.get_float("eps",0.01);
  std::vector<int> j    = pars.get_vints("j",2);
  std::vector<int> nlag = pars.get_vints("nlag",2);
  std::vector<int> nb   = pars.get_auxvints("nb",100,"in");
  /* Check PEF parameters */
  if(nlag.empty()) {
    seperr("Please supply either lag coordinates or the filter box dimensions\n");
  }
  if(nlag.size() != 2) {
    seperr("Please supply two nlag values or a list of lag coordinates\n");
  }
  if(j.empty()) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }
  if(j.size() != 2) {
    seperr("Please supply three j values\n");
  }

  /* Input data */
  sregf bldin = sregf("in");
  sregf indin = sregf("inddat");
  if((int)nb.size() != bldin.get_axis(3).n) {
    seperr("Number of blended shots and length of blending factor array is not consistent"
        " %d %d.\n",(int)nb.size(),indin.get_axis(3).n);
  }

  /* Data axes */
  int nt   = indin.get_axis(1).n; int nrx   = indin.get_axis(2).n;
  float ot = indin.get_axis(1).o; float orx = indin.get_axis(2).o;
  float dt = indin.get_axis(1).d; float drx = indin.get_axis(2).d;
  std::vector<int> nd;
  /* Padded axes */
  int ntp = pars.get_int("n1p",nt);
  int nrp = pars.get_int("n2p",nrx);
  nd.push_back(ntp); nd.push_back(nrp);
  if(ntp == nt || nrp == nrx) {
    seperr("You have not padded the time or receiver axes. Please do so and rerun.\n");
  }

  /* Build axes */
  std::vector<axis> saxes;
  saxes.push_back(bldin.get_axis(1));
  saxes.push_back(bldin.get_axis(2));
  saxes.push_back(bldin.get_axis(3));

  /* Build hypercubes */
  hypercube_float *blddat = new hypercube_float(bldin.get_axes(),true);
  hypercube_float *inddat = new hypercube_float(indin.get_axes(),true);
  hypercube_float *datsep = new hypercube_float(indin.get_axes(),true);

  bldin.read_all("in", blddat);
  indin.read_all("inddat", inddat);

  pefdblnd db = pefdblnd(pefniter, sniniter, eps, j, nlag, nb, saxes);
  db.set_padding(ntp, nrp);
  db.set_verb(vrb);
  db.set_nthrds(nthd);

  for(int itry = 0; itry < 1; ++itry) {
    db.apply(blddat, inddat, datsep);
  }

  /* Write out deblended data */
  sregf dout = sregf(datsep->get_axes());
  dout.write_description("out");
  dout.write_all("out", datsep);

  /* Clean up */
  delete blddat; delete inddat; delete datsep;
}
