/**
 * Tests the forward and adjoint of the
 * Born modeling operator for modeling
 * particle velocities in an acoustic
 * medium
 * @author: Joseph Jennings
 * @version: 2017.05.14
 */

#include "sep_params.h"
#include "../main/bbornpv.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int adj = pars.get_int("adj",0);
  int wfldj = pars.get_int("wfldwj",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  float rho = pars.get_float("rho",1000.0);
  /* Blending parameters */
  int minblnd  = pars.get_int("minblnd",2);
  int maxblnd  = pars.get_int("maxblnd",3);
  float maxdly = pars.get_float("maxdly",0.5);
  float mindly = pars.get_float("mindly",0.0);
  int space    = pars.get_int("space",7);
  int seed     = pars.get_int("seed",0);
  std::vector<float> stimes, xss;

  /* Input files */
  sregf datin; sregf modin;
  if(!adj) {
    modin = sregf("in");
  } else {
    datin = sregf("in");
  }
  sregf srcin = sregf("src");
  sregf velin = sregf("bvel");
  sregf bwfout; sregf awfout;

  /* Get source/receiver axes */
  int nrx; float orx; float drx;
  int nsx; float osx; float dsx;
  if(!adj) {
    nsx = pars.get_int  ("nsx",  1);
    osx = pars.get_float("osx",0.0);
    dsx = pars.get_float("dsx",1.0);
    nrx = pars.get_int  ("nrx",  1);
    orx = pars.get_float("orx",0.0);
    drx = pars.get_float("drx",1.0);
  } else {
    stimes = pars.get_auxvfloats("stimes",100,"in");
    xss    = pars.get_auxvfloats("xss",100,"in");
    if(xss.size() != stimes.size()) {
      seperr("number of shot times %d should = number of shot positions %d\n",(int)stimes.size(),(int)xss.size());
    }
    nrx = datin.get_axis(2).n;
    orx = datin.get_axis(2).o;
    drx = datin.get_axis(2).d;
  }

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  if(!adj) {
    if(!modin.get_axis(1).same_axis(velin.get_axis(1)) || ! modin.get_axis(2).same_axis(velin.get_axis(2))) {
      seperr("input model must have same axes as background velocity.\n");
    }
  } else {
    if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
      seperr("MigrateMCShots: input data and source wavelet must have same time axis.\n");
    if(datin.get_axis(3).n != 2)
      seperr("MigrateMCShots assumes that the component axis is the third axis.\n");
  }

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  float oz  = velin.get_axis(1).o;
  float dz  = velin.get_axis(1).d;
  int   nx  = velin.get_axis(2).n;
  float ox  = velin.get_axis(2).o;
  float dx  = velin.get_axis(2).d;
  /* Two component data */
  int nc = 2;

  /* Build axes */
  axis saxis;
  if(!adj) {
    saxis = axis(nsx,osx,dsx); //TODO: Need to change daxes so it has blended sources
  } else {
    saxis = axis((int)xss.size(),0.0,1.0);
  }
  std::vector<axis> daxes;
  daxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(nc,0.0,1.0));
  if(!adj) {
    daxes.push_back(saxis);
  }
  else {
    daxes.push_back(datin.get_axis(4));
  }

  std::vector<axis>paxes;
  paxes.push_back(axis(nz,oz,dz));
  paxes.push_back(axis(nx,ox,dx));
  paxes.push_back(srcin.get_axis(1));

  std::vector<axis> iaxes;
  if(adj) {
    iaxes.push_back(velin.get_axis(1));
    iaxes.push_back(velin.get_axis(2));
    iaxes.push_back(datin.get_axis(4));
  }

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);
  hypercube_float *mod = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  if(!adj) {
    modin.read_all("in",mod);
  } else {
    datin.read_all("in",dat);
  }
  srcin.read_all("src",src);
  velin.read_all("bvel",vel);

  /* Construct operator */
  bbornpv *bborn = nullptr;
  if(!stimes.empty()) {
    bborn = new bbornpv(paxes,daxes,srcz,recz,vel,src,rho,dtu,fmax,minblnd,maxblnd,stimes,xss);
  } else {
    bborn = new bbornpv(paxes,daxes,saxis,srcz,recz,vel,src,rho,dtu,fmax,minblnd,maxblnd,mindly,maxdly,seed,space);
  }
  if(!adj && !bwfname.empty()) {
    bwfout = sregf(paxes);
    bwfout.write_description(bwfname);
    bborn->setbwfldout("bwfldsout",wfldj);
  } else if(adj && !awfname.empty()) {
    awfout = sregf(paxes);
    awfout.write_description(awfname);
    bborn->setawfldout("awfldsout",wfldj);
  }

  if(!adj) {
    bborn->forward(false,mod,dat);
    sregf datout = sregf(dat->get_axes());
    datout.write_description("out");
    datout.write_all("out",dat);
    int nb = bborn->get_nbshots();
    putch("n4","d",&nb);
  } else {
    bborn->adjoint(false,mod,dat);
    sregf modout = sregf(mod->get_axes());
    modout.write_description("out");
    modout.write_all("out",mod);
    int one = 1;
    putch("n3","d",&one);
    putch("n4","d",&one);
  }

}
