/**
 * Tests forward-adjoint pair of the
 * linearized blended born modeling operator
 * @author: Joseph Jennings
 * @version: 2018.03.10
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "axis.h"
#include "hypercube_float.h"
#include "../main/bornp.h"
#include "../main/simpleplot.h"

class bornptest: public ::testing::Test {
  public:
    void pad(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad);
  protected:
    virtual void SetUp(){
    }

    virtual void TearDown() {
    }
};

TEST_F(bornptest,dptest) {

  int nx = 300; float ox = 0.0; float dx = 10.0;
  int nz = 300; float oz = 0.0; float dz = 10.0;
  int nt = 250; float ot = 0.0; float dts = 0.004; float dtu = 0.001;
  int nsx =  6; float osx = 60.0; float dsx = 10.0;
  int nrx = 40; float orx = 60.0; float drx = 1.0;
  int srcz = 60; int recz = 60;
  int p = 10; float fmax = 15.0;
  axis ax  = axis(nx,ox,dx);
  axis az  = axis(nz,oz,dz);
  axis axp = axis(nx+p,ox,dx);
  axis azp = axis(nz+p,oz,dz);
  axis at  = axis(nt,ot,dts);
  axis as  = axis(nsx,osx,dsx);
  axis ar  = axis(nrx,orx,drx);

  std::vector<axis> axes, paxes, uaxes, daxes, waxes;
  /* Velocity axes */
  axes.push_back(az);   axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavelet axes */
  waxes.push_back(at);
  /* Data axes */
  daxes.push_back(at); daxes.push_back(ar);
  daxes.push_back(as);
  /* Wavefield axes */
  uaxes.push_back(azp); uaxes.push_back(axp);
  uaxes.push_back(at);

  hypercube_float *vel  = new hypercube_float(axes,true);
  hypercube_float *pvel = new hypercube_float(paxes,true);
  hypercube_float *src  = new hypercube_float(waxes,true);

  vel->set_val(3000.0); src->random(); src->scale(0.000001);
  pad(p/2,p/2,vel,pvel);

  hypercube_float *rm = new hypercube_float(axes,true);
  hypercube_float *m  = new hypercube_float(paxes,true);
  hypercube_float *ms = new hypercube_float(paxes,true);
  hypercube_float *d  = new hypercube_float(daxes,true);
  hypercube_float *ds = new hypercube_float(daxes,true);

  rm->random(); d->random();
  pad(p/2,p/2,rm,m);

  bornp born = bornp(uaxes, daxes, as, srcz, recz, pvel, src, dtu, fmax);

  /* Add == false */
  born.forward(false,m,ds);
  double dotd = d->dot(ds);
  born.adjoint(false,ms,d);
  double dotm = m->dot(ms);
  fprintf(stderr,"%f %f\n",dotd,dotm);
  ASSERT_NEAR(dotd,dotm,0.0001);

  /* Clean up */
  delete vel; delete pvel; delete src;
  delete rm; delete m; delete ms;
  delete d; delete ds;
}

/**
 * Pads a 2D image
 * @param loc1 the offset along axis 1
 * @param loc2 the offset along axis 1
 * @param dat  the input data
 * @param pad  the padded result
 */
void bornptest::pad(int loc1, int loc2, hypercube_float *dat, hypercube_float* pad) {
  pad->zero();
  int n1p = pad->get_axis(1).n;
  int n1  = dat->get_axis(1).n; int n2  = dat->get_axis(2).n;
  for(int i2 = 0; i2 < n2; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      pad->vals[(i2+loc2)*n1p + i1 + loc1] = dat->vals[i2*n1 + i1];
    }
  }
}
