/**
 * Tests the forward and adjoint of the
 * Born modeling operator for modeling
 * linearized pressure data in an acoustic medium
 * @author: Joseph Jennings
 * @version: 2018.03.17
 */

#include "sep_params.h"
#include "../main/bornp.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int adj = pars.get_int("adj",0);
  int wfldj = pars.get_int("wfldwj",10);
  int nthd  = pars.get_int("nthrds",1);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);

  /* Input files */
  sregf datin; sregf modin;
  if(!adj) {
    modin = sregf("in");
  } else {
    datin = sregf("in");
  }
  sregf srcin = sregf("src");
  sregf velin = sregf("bvel");
  sregf bwfout; sregf awfout;

  /* Get source/receiver axes */
  int nrx; float orx; float drx;
  int nsx = 1; float osx = 0.0; float dsx = 1.0;
  if(!adj) {
    nsx = pars.get_int  ("nsx",  1);
    osx = pars.get_float("osx",0.0);
    dsx = pars.get_float("dsx",1.0);
    nrx = pars.get_int  ("nrx",  1);
    orx = pars.get_float("orx",0.0);
    drx = pars.get_float("drx",1.0);
  } else {
    nrx = datin.get_axis(2).n;
    orx = datin.get_axis(2).o;
    drx = datin.get_axis(2).d;
  }

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  if(!adj) {
    if(!modin.get_axis(1).same_axis(velin.get_axis(1)) || ! modin.get_axis(2).same_axis(velin.get_axis(2))) {
      seperr("input model must have same axes as background velocity.\n");
    }
  } else {
    if(!datin.get_axis(1).same_axis(srcin.get_axis(1)))
      seperr("bornp: input data and source wavelet must have same time axis.\n");
  }

  /* Velocity axes */
  int   nz  = velin.get_axis(1).n;
  float oz  = velin.get_axis(1).o;
  float dz  = velin.get_axis(1).d;
  int   nx  = velin.get_axis(2).n;
  float ox  = velin.get_axis(2).o;
  float dx  = velin.get_axis(2).d;

  /* Build axes */
  axis saxis;
  if(!adj) {
    saxis = axis(nsx,osx,dsx);
  } else {
    saxis = datin.get_axis(3);
  }

  std::vector<axis> daxes;
  daxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx));
  if(!adj) {
    daxes.push_back(saxis);
  }
  else {
    daxes.push_back(datin.get_axis(3));
  }

  std::vector<axis>paxes;
  paxes.push_back(axis(nz,oz,dz));
  paxes.push_back(axis(nx,ox,dx));
  paxes.push_back(srcin.get_axis(1));

  std::vector<axis> iaxes;
  if(adj) {
    iaxes.push_back(velin.get_axis(1));
    iaxes.push_back(velin.get_axis(2));
    iaxes.push_back(datin.get_axis(3));
  }

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);
  hypercube_float *mod = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  if(!adj) {
    modin.read_all("in",mod);
  } else {
    datin.read_all("in",dat);
  }
  srcin.read_all("src",src);
  velin.read_all("bvel",vel);

  /* Construct operator */
  bornp *born = new bornp(paxes,daxes,saxis,srcz,recz,vel,src,dtu,fmax);
  born->setnthrds(nthd);

  if(!adj && !bwfname.empty()) {
    bwfout = sregf(paxes);
    bwfout.write_description(bwfname);
    born->setbwfldout("bwfldsout",wfldj);
  } else if(adj && !awfname.empty()) {
    awfout = sregf(paxes);
    awfout.write_description(awfname);
    born->setawfldout("awfldsout",wfldj);
  }

  if(!adj) {
    born->forward(false,mod,dat);
    sregf datout = sregf(dat->get_axes());
    datout.write_description("out");
    datout.write_all("out",dat);
    //putch("n4","d",&nb);
  } else {
    born->adjoint(false,mod,dat);
    sregf modout = sregf(mod->get_axes());
    modout.write_description("out");
    modout.write_all("out",mod);
    int one = 1;
    putch("n3","d",&one);
    putch("n4","d",&one);
  }

  /* Clean up */
  delete vel; delete src;
  delete dat; delete mod;

}
