/**
 * Tests the large 2D PEF deblending operator
 * forward and adjoint
 * @author: Joseph Jennings
 * @version: 2017.04.29
 */

#include <limits.h>
#include <random>
#include "gtest/gtest.h"
#include "axis.h"
#include "hypercube_float.h"
#include "super_vector.h"
#include "../main/pefsignoi2d.h"
#include "../main/simpleplot.h"

class pefsignoi2dtest: public ::testing::Test {
  protected:
    virtual void SetUp(){
    }

    virtual void TearDown() {
    }
};

TEST_F(pefsignoi2dtest,forwardtot) {

  std::vector<int> nd, j, nlag;
  nd.push_back(100); nd.push_back(100);
  j.push_back(10); j.push_back(10);
  nlag.push_back(4); nlag.push_back(3);

  /* Build signal and noise PEFs */
  std::vector<axis> daxes,faxes;
  faxes.push_back(axis(10,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  hypercube_float *auxs = new hypercube_float(faxes,true);
  hypercube_float *auxn = new hypercube_float(faxes,true);
  auxs->random(); auxn->random();

  lv_conv2dm *lv2s = new lv_conv2dm(nd,j,nlag,auxs);
  lv_conv2dm *lv2n = new lv_conv2dm(nd,j,nlag,auxn);

  /* Model and data vector */
  daxes.push_back(axis(nd[0],0.0,1.0));
  daxes.push_back(axis(nd[1],0.0,1.0));
  hypercube_float *m1  = new hypercube_float(daxes,true);
  hypercube_float *m2  = new hypercube_float(daxes,true);
  m1->random(); m2->random();
  super_vector *m      = new super_vector(m1,m2);
  hypercube_float *d   = new hypercube_float(daxes,true);
  hypercube_float *dn  = new hypercube_float(daxes,true);
  hypercube_float *ds  = new hypercube_float(daxes,true);
  super_vector *wdat   = new super_vector(dn,ds);
  super_vector *datout = new super_vector(d,wdat);

  pefsignoi2d A = pefsignoi2d(d,lv2s,lv2n);

  /* Build total operator */
  A.forward(false,m,datout);

  /* Check the blended data (gamma) */
  hypercube_float *obld = new hypercube_float(daxes,true);
  obld->scale_add(1.0,m1,1.0,m2);
  ASSERT_NEAR(obld->sum(),datout->vecs[0]->sum(),0.000001);

  /* The noise PEF applied to the signal */
  lv2n->forward(false,m1,d);
  ASSERT_NEAR(d->sum(),((super_vector*)datout->vecs[1])->vecs[0]->sum(),0.000001);

  /* The signal PEF applied to the noise */
  lv2s->forward(false,m2,d);
  ASSERT_NEAR(d->sum(),((super_vector*)datout->vecs[1])->vecs[1]->sum(),0.000001);
}

TEST_F(pefsignoi2dtest,adjointtot) {

  std::vector<int> nd, j, nlag;
  nd.push_back(100); nd.push_back(100);
  j.push_back(10); j.push_back(10);
  nlag.push_back(4); nlag.push_back(3);

  /* Build signal and noise PEFs */
  std::vector<axis> daxes,faxes;
  faxes.push_back(axis(10,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  hypercube_float *auxs = new hypercube_float(faxes,true);
  hypercube_float *auxn = new hypercube_float(faxes,true);
  auxs->random(); auxn->random();
  lv_conv2dm *lv2s = new lv_conv2dm(nd,j,nlag,auxs);
  lv_conv2dm *lv2n = new lv_conv2dm(nd,j,nlag,auxn);

  /* Model and data vector */
  daxes.push_back(axis(nd[0],0.0,1.0));
  daxes.push_back(axis(nd[1],0.0,1.0));
  hypercube_float *m1  = new hypercube_float(daxes,true);
  hypercube_float *m2  = new hypercube_float(daxes,true);
  super_vector *m      = new super_vector(m1,m2);
  hypercube_float *d   = new hypercube_float(daxes,true);
  hypercube_float *dn  = new hypercube_float(daxes,true);
  hypercube_float *ds  = new hypercube_float(daxes,true);
  d->random(); dn->random(); ds->random();
  super_vector *wdat   = new super_vector(dn,ds);
  super_vector *datout = new super_vector(d,wdat);

  pefsignoi2d A = pefsignoi2d(d,lv2s,lv2n);

  /* Build total operator */
  A.adjoint(false,m,datout);

  hypercube_float *t1 = new hypercube_float(daxes,true);
  hypercube_float *t2 = new hypercube_float(daxes,true);
  /* Check the signal update */
  lv2n->adjoint(false,t1,dn);
  t2->scale_add(1.0,d,1.0,t1);
  fprintf(stderr,"%f %f\n",t2->sum(),m->vecs[0]->sum());
  ASSERT_NEAR(t2->sum(),m->vecs[0]->sum(),0.0001);

  /* Check the noise update */
  lv2s->adjoint(false,t1,ds);
  t2->scale_add(1.0,d,1.0,t1);
  fprintf(stderr,"%f %f\n",t2->sum(),m->vecs[1]->sum());
  ASSERT_NEAR(t2->sum(),m->vecs[1]->sum(),0.0001);
}

TEST_F(pefsignoi2dtest,dptest) {

  std::vector<int> nd, j, nlag;
  nd.push_back(100); nd.push_back(100);
  j.push_back(10); j.push_back(10);
  nlag.push_back(4); nlag.push_back(3);

  /* Build signal and noise PEFs */
  std::vector<axis> daxes,faxes;
  faxes.push_back(axis(10,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  faxes.push_back(axis(11,0.0,1.0));
  hypercube_float *auxs = new hypercube_float(faxes,true);
  hypercube_float *auxn = new hypercube_float(faxes,true);
  auxs->random(); auxn->random();
  lv_conv2dm *lv2s = new lv_conv2dm(nd,j,nlag,auxs);
  lv_conv2dm *lv2n = new lv_conv2dm(nd,j,nlag,auxn);

  /* Model and data vector */
  daxes.push_back(axis(nd[0],0.0,1.0));
  daxes.push_back(axis(nd[1],0.0,1.0));
  hypercube_float *m1  = new hypercube_float(daxes,true);
  hypercube_float *m2  = new hypercube_float(daxes,true);
  m1->random(); m2->random();
  super_vector *m      = new super_vector(m1,m2);
  hypercube_float *d   = new hypercube_float(daxes,true);
  hypercube_float *dn  = new hypercube_float(daxes,true);
  hypercube_float *ds  = new hypercube_float(daxes,true);
  d->random(); dn->random(); ds->random();
  super_vector *wdat   = new super_vector(dn,ds);
  super_vector *datout = new super_vector(d,wdat);

  /* Temporary vectors for dot products */
  hypercube_float *mt1 = new hypercube_float(daxes,true);
  hypercube_float *mt2 = new hypercube_float(daxes,true);
  super_vector *mt = new super_vector(mt1,mt2);
  hypercube_float *dt1 = new hypercube_float(daxes,true);
  hypercube_float *dt2 = new hypercube_float(daxes,true);
  hypercube_float *dtd = new hypercube_float(daxes,true);
  super_vector *wdatt = new super_vector(dt1,dt2);
  super_vector *dt = new super_vector(dtd,wdatt);

  pefsignoi2d A = pefsignoi2d(d,lv2s,lv2n);

  /* Dot product test */
  A.forward(false,m,dt);
  double dotd = dt->dot(datout);
  A.adjoint(false,mt,datout);
  double dotm = mt->dot(m);
  fprintf(stderr,"%f %f\n",dotd,dotm);
  ASSERT_NEAR(dotm,dotd,0.0001);
}
